@@ -1,13 +1,13 @@
-/* $Id: pstream.h,v 1.84 2004/10/17 15:43:16 redi Exp $
+/* $Id: pstream.h,v 1.112 2010/03/20 14:50:47 redi Exp $
 PStreams - POSIX Process I/O for C++
-Copyright (C) 2001,2002,2003,2004 Jonathan Wakely
+Copyright (C) 2001,2002,2003,2004,2005,2006,2007,2008 Jonathan Wakely
 
 This file is part of PStreams.
 
 PStreams is free software; you can redistribute it and/or modify
-it under the terms of the GNU Lesser General Public License as
-published by the Free Software Foundation; either version 2.1 of
-the License, or (at your option) any later version.
+it under the terms of the GNU Lesser General Public License as published by
+the Free Software Foundation; either version 3 of the License, or
+(at your option) any later version.
 
 PStreams is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -15,8 +15,7 @@ MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.
 
 You should have received a copy of the GNU Lesser General Public License
-along with PStreams; if not, write to the Free Software Foundation, Inc.,
-59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
 /**
@@ -38,27 +37,31 @@ along with PStreams; if not, write to the Free Software Foundation, Inc.,
 #include <string>
 #include <vector>
 #include <algorithm>    // for min()
-#include <cstring>      // for memcpy(), memmove() etc.
 #include <cerrno>       // for errno
 #include <cstddef>      // for size_t
 #include <cstdlib>      // for exit()
 #include <sys/types.h>  // for pid_t
 #include <sys/wait.h>   // for waitpid()
+#include <sys/ioctl.h>  // for ioctl() and FIONREAD
+#if defined(__sun)
+# include <sys/filio.h> // for FIONREAD on Solaris 2.5
+#endif
 #include <unistd.h>     // for pipe() fork() exec() and filedes functions
 #include <signal.h>     // for kill()
+#include <fcntl.h>      // for fcntl()
 #if REDI_EVISCERATE_PSTREAMS
-#include <stdio.h>       // for FILE, fdopen()
+# include <stdio.h>     // for FILE, fdopen()
 #endif
 
 
 /// The library version.
-#define PSTREAMS_VERSION 0x0050   // 0.5.0
+#define PSTREAMS_VERSION 0x0070   // 0.7.0
 
 /**
  *  @namespace redi
  *  @brief  All PStreams classes are declared in namespace redi.
  *
- *  Like the standard IOStreams, PStreams is a set of class templates,
+ *  Like the standard iostreams, PStreams is a set of class templates,
  *  taking a character type and traits type. As with the standard streams
  *  they are most likely to be used with @c char and the default
  *  traits type, so typedefs for this most common case are provided.
@@ -78,6 +81,9 @@ namespace redi
     /// Type used to hold the arguments for a command.
     typedef std::vector<std::string>          argv_type;
 
+    /// Type used for file descriptors.
+    typedef int                               fd_type;
+
     static const pmode pstdin  = std::ios_base::out; ///< Write to stdin
     static const pmode pstdout = std::ios_base::in;  ///< Read from stdout
     static const pmode pstderr = std::ios_base::app; ///< Read from stderr
@@ -100,9 +106,7 @@ namespace redi
       typedef typename traits_type::int_type    int_type;
       typedef typename traits_type::off_type    off_type;
       typedef typename traits_type::pos_type    pos_type;
-      /// Type used for file descriptors.
-      typedef int                               fd_type;
-      /** @deprecated use fd_type instead. */
+      /** @deprecated use pstreams::fd_type instead. */
       typedef fd_type                           fd_t;
 
       /// Default constructor.
@@ -154,14 +158,14 @@ namespace redi
 #if REDI_EVISCERATE_PSTREAMS
       /// Obtain FILE pointers for each of the process' standard streams.
       std::size_t
-      fopen(std::FILE*& in, std::FILE*& out, std::FILE*& err);
+      fopen(FILE*& in, FILE*& out, FILE*& err);
 #endif
 
       /// Return the exit status of the process.
       int
       status() const;
 
-      /// Return the error number for the most recent failed operation.
+      /// Return the error number (errno) for the most recent failed operation.
       int
       error() const;
 
@@ -188,12 +192,16 @@ namespace redi
 
       /// Insert a sequence of characters into the pipe.
       std::streamsize
-      write(char_type* s, std::streamsize n);
+      write(const char_type* s, std::streamsize n);
 
       /// Extract a sequence of characters from the pipe.
       std::streamsize
       read(char_type* s, std::streamsize n);
 
+      /// Report how many characters can be read from active input without blocking.
+      std::streamsize
+      showmanyc();
+
     protected:
       /// Enumerated type to indicate whether stdout or stderr is to be read.
       enum buf_read_src { rsrc_out = 0, rsrc_err = 1 };
@@ -229,7 +237,7 @@ namespace redi
       empty_buffer();
 
       bool
-      fill_buffer();
+      fill_buffer(bool non_blocking = false);
 
       /// Return the active input buffer.
       char_type*
@@ -266,6 +274,9 @@ namespace redi
     protected:
       typedef basic_pstreambuf<CharT, Traits>       streambuf_type;
 
+      typedef pstreams::pmode                       pmode;
+      typedef pstreams::argv_type                   argv_type;
+
       /// Default constructor.
       pstream_common();
 
@@ -273,9 +284,7 @@ namespace redi
       pstream_common(const std::string& command, pmode mode);
 
       /// Constructor that initialises the stream by starting a process.
-      pstream_common( const std::string& file,
-                      const argv_type& argv,
-                      pmode mode );
+      pstream_common(const std::string& file, const argv_type& argv, pmode mode);
 
       /// Pure virtual destructor.
       virtual
@@ -309,7 +318,7 @@ namespace redi
 #if REDI_EVISCERATE_PSTREAMS
       /// Obtain FILE pointers for each of the process' standard streams.
       std::size_t
-      fopen(std::FILE*& in, std::FILE*& out, std::FILE*& err);
+      fopen(FILE*& in, FILE*& out, FILE*& err);
 #endif
 
     protected:
@@ -339,6 +348,13 @@ namespace redi
 
       using pbase_type::buf_;  // declare name in this scope
 
+      pmode readable(pmode mode)
+      {
+        if (!(mode & (pstdout|pstderr)))
+          mode |= pstdout;
+        return mode;
+      }
+
     public:
       /// Type used to specify how to connect to the process.
       typedef typename pbase_type::pmode            pmode;
@@ -362,7 +378,7 @@ namespace redi
        * @see   do_open(const std::string&, pmode)
        */
       basic_ipstream(const std::string& command, pmode mode = pstdout)
-      : istream_type(NULL), pbase_type(command, mode|pstdout)
+      : istream_type(NULL), pbase_type(command, readable(mode))
       { }
 
       /**
@@ -379,7 +395,7 @@ namespace redi
       basic_ipstream( const std::string& file,
                       const argv_type& argv,
                       pmode mode = pstdout )
-      : istream_type(NULL), pbase_type(file, argv, mode|pstdout)
+      : istream_type(NULL), pbase_type(file, argv, readable(mode))
       { }
 
       /**
@@ -402,7 +418,7 @@ namespace redi
       void
       open(const std::string& command, pmode mode = pstdout)
       {
-        this->do_open(command, mode|pstdout);
+        this->do_open(command, readable(mode));
       }
 
       /**
@@ -420,7 +436,7 @@ namespace redi
             const argv_type& argv,
             pmode mode = pstdout )
       {
-        this->do_open(file, argv, mode|pstdout);
+        this->do_open(file, argv, readable(mode));
       }
 
       /**
@@ -598,8 +614,7 @@ namespace redi
        * @param mode     the I/O mode to use when opening the pipe.
        * @see   do_open(const std::string&, pmode)
        */
-      basic_pstream( const std::string& command,
-                     pmode mode = pstdout|pstdin )
+      basic_pstream(const std::string& command, pmode mode = pstdout|pstdin)
       : iostream_type(NULL), pbase_type(command, mode)
       { }
 
@@ -727,7 +742,7 @@ namespace redi
 
       /// Default constructor, creates an uninitialised stream.
       basic_rpstream()
-      : ostream_type(NULL) , istream_type(NULL) , pbase_type()
+      : ostream_type(NULL), istream_type(NULL), pbase_type()
       { }
 
       /**
@@ -838,7 +853,7 @@ namespace redi
   /**
    * When inserted into an output pstream the manipulator calls
    * basic_pstreambuf<C,T>::peof() to close the output pipe,
-   * causing the child process to receive the End Of File indicator
+   * causing the child process to receive the end-of-file indicator
    * on subsequent reads from its @c stdin stream.
    *
    * @brief   Manipulator to close the pipe connected to the process' stdin.
@@ -941,72 +956,133 @@ namespace redi
     }
 
   /**
-   * Starts a new process by passing @a command to the shell
+   * Starts a new process by passing @a command to the shell (/bin/sh)
    * and opens pipes to the process with the specified @a mode.
    *
+   * If @a mode contains @c pstdout the initial read source will be
+   * the child process' stdout, otherwise if @a mode  contains @c pstderr
+   * the initial read source will be the child's stderr.
+   *
+   * Will duplicate the actions of  the  shell  in searching for an
+   * executable file if the specified file name does not contain a slash (/)
+   * character.
+   *
+   * @warning
    * There is no way to tell whether the shell command succeeded, this
    * function will always succeed unless resource limits (such as
    * memory usage, or number of processes or open files) are exceeded.
+   * This means is_open() will return true even if @a command cannot
+   * be executed.
+   * Use pstreambuf::open(const std::string&, const argv_type&, pmode)
+   * if you need to know whether the command failed to execute.
    *
    * @param   command  a string containing a shell command.
    * @param   mode     a bitwise OR of one or more of @c out, @c in, @c err.
    * @return  NULL if the shell could not be started or the
    *          pipes could not be opened, @c this otherwise.
-   * @see     execlp(3)
+   * @see     <b>execl</b>(3)
    */
   template <typename C, typename T>
     basic_pstreambuf<C,T>*
     basic_pstreambuf<C,T>::open(const std::string& command, pmode mode)
     {
+      const char * shell_path = "/bin/sh";
+#if 0
+      const std::string argv[] = { "sh", "-c", command };
+      return this->open(shell_path, argv_type(argv, argv+3), mode);
+#else
       basic_pstreambuf<C,T>* ret = NULL;
 
       if (!is_open())
       {
         switch(fork(mode))
         {
-          case 0 :
-          {
-            // this is the new process, exec command
-            ::execlp("sh", "sh", "-c", command.c_str(), (void*)NULL);
+        case 0 :
+          // this is the new process, exec command
+          ::execl(shell_path, "sh", "-c", command.c_str(), (char*)NULL);
 
-            // can only reach this point if exec() failed
+          // can only reach this point if exec() failed
 
-            // parent can get exit code from waitpid()
-            ::_exit(errno);
-            // using std::exit() would make static dtors run twice
-          }
-          case -1 :
-          {
-            // couldn't fork, error already handled in pstreambuf::fork()
-            break;
-          }
-          default :
-          {
-            // this is the parent process
-            // activate buffers
-            create_buffers(mode);
-            ret = this;
-          }
+          // parent can get exit code from waitpid()
+          ::_exit(errno);
+          // using std::exit() would make static dtors run twice
+
+        case -1 :
+          // couldn't fork, error already handled in pstreambuf::fork()
+          break;
+
+        default :
+          // this is the parent process
+          // activate buffers
+          create_buffers(mode);
+          ret = this;
         }
       }
       return ret;
+#endif
+    }
+
+  /**
+   * @brief  Helper function to close a file descriptor.
+   *
+   * Inspects @a fd and calls <b>close</b>(3) if it has a non-negative value.
+   *
+   * @param   fd  a file descriptor.
+   * @relates basic_pstreambuf
+   */
+  inline void
+  close_fd(pstreams::fd_type& fd)
+  {
+    if (fd >= 0 && ::close(fd) == 0)
+      fd = -1;
+  }
+
+  /**
+   * @brief  Helper function to close an array of file descriptors.
+   *
+   * Calls @c close_fd() on each member of the array.
+   * The length of the array is determined automatically by
+   * template argument deduction to avoid errors.
+   *
+   * @param   fds  an array of file descriptors.
+   * @relates basic_pstreambuf
+   */
+  template <int N>
+    inline void
+    close_fd_array(pstreams::fd_type (&fds)[N])
+    {
+      for (std::size_t i = 0; i < N; ++i)
+        close_fd(fds[i]);
     }
 
   /**
    * Starts a new process by executing @a file with the arguments in
    * @a argv and opens pipes to the process with the specified @a mode.
    *
-   * By convention argv[0] should be the file name of the file being executed.
+   * By convention @c argv[0] should be the file name of the file being
+   * executed.
+   *
+   * If @a mode contains @c pstdout the initial read source will be
+   * the child process' stdout, otherwise if @a mode  contains @c pstderr
+   * the initial read source will be the child's stderr.
+   *
    * Will duplicate the actions of  the  shell  in searching for an
    * executable file if the specified file name does not contain a slash (/)
    * character.
    *
+   * Iff @a file is successfully executed then is_open() will return true.
+   * Otherwise, pstreambuf::error() can be used to obtain the value of
+   * @c errno that was set by <b>execvp</b>(3) in the child process.
+   *
+   * The exit status of the new process will be returned by
+   * pstreambuf::status() after pstreambuf::exited() returns true.
+   *
    * @param   file  a string containing the pathname of a program to execute.
    * @param   argv  a vector of argument strings passed to the new program.
    * @param   mode  a bitwise OR of one or more of @c out, @c in and @c err.
    * @return  NULL if a pipe could not be opened or if the program could
    *          not be executed, @c this otherwise.
-   * @see     execvp(3)
+   * @see     <b>execvp</b>(3)
    */
   template <typename C, typename T>
     basic_pstreambuf<C,T>*
@@ -1018,72 +1094,90 @@ namespace redi
 
       if (!is_open())
       {
-        switch(fork(mode))
+        // constants for read/write ends of pipe
+        enum { RD, WR };
+
+        // open another pipe and set close-on-exec
+        fd_type ck_exec[] = { -1, -1 };
+        if (-1 == ::pipe(ck_exec)
+            || -1 == ::fcntl(ck_exec[RD], F_SETFD, FD_CLOEXEC)
+            || -1 == ::fcntl(ck_exec[WR], F_SETFD, FD_CLOEXEC))
         {
+          error_ = errno;
+          close_fd_array(ck_exec);
+        }
+        else
+        {
+          switch(fork(mode))
+          {
           case 0 :
-          {
             // this is the new process, exec command
-
-            char** arg_v = new char*[argv.size()+1];
-            for (std::size_t i = 0; i < argv.size(); ++i)
             {
-              const std::string& src = argv[i];
-              char*& dest = arg_v[i];
-              dest = new char[src.size()+1];
-              dest[ src.copy(dest, src.size()) ] = '\0';
+              char** arg_v = new char*[argv.size()+1];
+              for (std::size_t i = 0; i < argv.size(); ++i)
+              {
+                const std::string& src = argv[i];
+                char*& dest = arg_v[i];
+                dest = new char[src.size()+1];
+                dest[ src.copy(dest, src.size()) ] = '\0';
+              }
+              arg_v[argv.size()] = NULL;
+
+              ::execvp(file.c_str(), arg_v);
+
+              // can only reach this point if exec() failed
+
+              // parent can get error code from ck_exec pipe
+              error_ = errno;
+
+              ::write(ck_exec[WR], &error_, sizeof(error_));
+              ::close(ck_exec[WR]);
+              ::close(ck_exec[RD]);
+
+              ::_exit(error_);
+              // using std::exit() would make static dtors run twice
             }
-            arg_v[argv.size()] = NULL;
 
-            ::execvp(file.c_str(), arg_v);
-
-            // can only reach this point if exec() failed
-
-            // parent can get exit code from waitpid()
-            ::_exit(errno);
-            // using std::exit() would make static dtors run twice
-          }
           case -1 :
-          {
             // couldn't fork, error already handled in pstreambuf::fork()
+            close_fd_array(ck_exec);
             break;
-          }
+
           default :
-          {
             // this is the parent process
-            // activate buffers
-            create_buffers(mode);
-            ret = this;
+
+            // check child called exec() successfully
+            ::close(ck_exec[WR]);
+            switch (::read(ck_exec[RD], &error_, sizeof(error_)))
+            {
+            case 0:
+              // activate buffers
+              create_buffers(mode);
+              ret = this;
+              break;
+            case -1:
+              error_ = errno;
+              break;
+            default:
+              // error_ contains error code from child
+              // call wait() to clean up and set ppid_ to 0
+              this->wait();
+              break;
+            }
+            ::close(ck_exec[RD]);
           }
         }
       }
       return ret;
     }
 
-  /**
-   * @brief  Helper function to close an array of file descriptors.
-   *
-   * Inspects each of the @a count file descriptors in the array @a filedes
-   * and calls @c close() if they have a non-negative value.
-   * @param filedes an array of file descriptors
-   * @param count size of the array.
-   * @relates basic_pstreambuf
-   */
-  inline void
-  close_fd_array(int* filedes, std::size_t count)
-  {
-    for (std::size_t i = 0; i < count; ++i)
-      if (filedes[i] >= 0)
-        if (::close(filedes[i]) == 0)
-          filedes[i] = -1;
-  }
-
   /**
    * Creates pipes as specified by @a mode and calls @c fork() to create
    * a new process. If the fork is successful the parent process stores
    * the child's PID and the opened pipes and the child process replaces
    * its standard streams with the opened pipes.
    *
-   * If an error occurs the error code will be set to one of the possile
+   * If an error occurs the error code will be set to one of the possible
    * errors for @c pipe() or @c fork().
    * See your system's documentation for these error codes.
    *
@@ -1099,17 +1193,16 @@ namespace redi
     {
       pid_t pid = -1;
 
-      // three pairs of file descriptors, for pipes connected to the
+      // Three pairs of file descriptors, for pipes connected to the
       // process' stdin, stdout and stderr
       // (stored in a single array so close_fd_array() can close all at once)
-      fd_type fd[6] =  {-1, -1, -1, -1, -1, -1};
+      fd_type fd[] = { -1, -1, -1, -1, -1, -1 };
       fd_type* const pin = fd;
       fd_type* const pout = fd+2;
       fd_type* const perr = fd+4;
 
       // constants for read/write ends of pipe
-      const int RD = 0;
-      const int WR = 1;
+      enum { RD, WR };
 
       // N.B.
       // For the pstreambuf pin is an output stream and
@@ -1161,7 +1254,7 @@ namespace redi
             // couldn't fork for some reason
             error_ = errno;
             // close any open pipes
-            close_fd_array(fd, 6);
+            close_fd_array(fd);
             break;
           }
           default :
@@ -1185,19 +1278,13 @@ namespace redi
               rpipe_[rsrc_err] = perr[RD];
               ::close(perr[WR]);
             }
-
-            if (rpipe_[rsrc_out] == -1 && rpipe_[rsrc_err] >= 0)
-            {
-              // reading stderr but not stdout, so use stderr for all reads
-              read_err(true);
-            }
           }
         }
       }
       else
       {
         // close any pipes we opened before failure
-        close_fd_array(fd, 6);
+        close_fd_array(fd);
       }
       return pid;
     }
@@ -1215,23 +1302,25 @@ namespace redi
     basic_pstreambuf<C,T>*
     basic_pstreambuf<C,T>::close()
     {
-      basic_pstreambuf<C,T>* ret = NULL;
-      if (is_open())
+      const bool running = is_open();
+
+      sync(); // this might call wait() and reap the child process
+
+      // rather than trying to work out whether or not we need to clean up
+      // just do it anyway, all cleanup functions are safe to call twice.
+
+      destroy_buffers(pstdin|pstdout|pstderr);
+
+      // close pipes before wait() so child gets EOF/SIGPIPE
+      close_fd(wpipe_);
+      close_fd_array(rpipe_);
+
+      do
       {
-        sync();
+        error_ = 0;
+      } while (wait() == -1 && error() == EINTR);
 
-        destroy_buffers(pstdin|pstdout|pstderr);
-
-        // close pipes before wait() so child gets EOF/SIGPIPE
-        close_fd_array(&wpipe_, 1);
-        close_fd_array(rpipe_, 2);
-
-        if (wait() == 1)
-        {
-          ret = this;
-        }
-      }
-      return ret;
+      return running ? this : NULL;
     }
 
   /**
@@ -1260,17 +1349,20 @@ namespace redi
       {
         delete[] rbuffer_[rsrc_out];
         rbuffer_[rsrc_out] = new char_type[bufsz];
-        if (rsrc_ == rsrc_out)
-          this->setg(rbuffer_[rsrc_out] + pbsz, rbuffer_[rsrc_out] + pbsz,
-              rbuffer_[rsrc_out] + pbsz);
+        rsrc_ = rsrc_out;
+        this->setg(rbuffer_[rsrc_out] + pbsz, rbuffer_[rsrc_out] + pbsz,
+            rbuffer_[rsrc_out] + pbsz);
       }
       if (mode & pstderr)
       {
         delete[] rbuffer_[rsrc_err];
         rbuffer_[rsrc_err] = new char_type[bufsz];
-        if (rsrc_ == rsrc_err)
+        if (!(mode & pstdout))
+        {
+          rsrc_ = rsrc_err;
           this->setg(rbuffer_[rsrc_err] + pbsz, rbuffer_[rsrc_err] + pbsz,
               rbuffer_[rsrc_err] + pbsz);
+        }
       }
     }
 
@@ -1319,10 +1411,14 @@ namespace redi
    * Suspends execution and waits for the associated process to exit, or
    * until a signal is delivered whose action is to terminate the current
    * process or to call a signal handling function. If the process has
-   * already exited wait() returns immediately.
+   * already exited (i.e. it is a "zombie" process) then wait() returns
+   * immediately.  Waiting for the child process causes all its system
+   * resources to be freed.
+   *
+   * error() will return EINTR if wait() is interrupted by a signal.
    *
    * @param   nohang  true to return immediately if the process has not exited.
-   * @return  1 if the process has exited.
+   * @return  1 if the process has exited and wait() has not yet been called.
    *          0 if @a nohang is true and the process has not exited yet.
    *          -1 if no process has been started or if an error occurs,
    *          in which case the error can be found using error().
@@ -1349,9 +1445,11 @@ namespace redi
             ppid_ = 0;
             status_ = status;
             exited = 1;
-            destroy_buffers(pstdin|pstdout|pstderr);
-            close_fd_array(&wpipe_, 1);
-            close_fd_array(rpipe_, 2);
+            // Close wpipe, would get SIGPIPE if we used it.
+            destroy_buffers(pstdin);
+            close_fd(wpipe_);
+            // Must free read buffers and pipes on destruction
+            // or next call to open()/close()
             break;
         }
       }
@@ -1379,7 +1477,11 @@ namespace redi
           error_ = errno;
         else
         {
+#if 0
           // TODO call exited() to check for exit and clean up? leave to user?
+          if (signal==SIGTERM || signal==SIGKILL)
+            this->exited();
+#endif
           ret = this;
         }
       }
@@ -1387,8 +1489,11 @@ namespace redi
     }
 
   /**
+   *  This function can call pstreambuf::wait() and so may change the
+   *  object's state if the child process has already exited.
+   *
    *  @return  True if the associated process has exited, false otherwise.
-   *  @see     basic_pstreambuf<C,T>::close()
+   *  @see     basic_pstreambuf<C,T>::wait()
    */
   template <typename C, typename T>
     inline bool
@@ -1399,9 +1504,9 @@ namespace redi
 
 
   /**
-   *  @return  The exit status of the child process, or -1 if close()
+   *  @return  The exit status of the child process, or -1 if wait()
    *           has not yet been called to wait for the child to exit.
-   *  @see     basic_pstreambuf<C,T>::close()
+   *  @see     basic_pstreambuf<C,T>::wait()
    */
   template <typename C, typename T>
     inline int
@@ -1422,7 +1527,7 @@ namespace redi
 
   /**
    *  Closes the output pipe, causing the child process to receive the
-   *  End Of File indicator on subsequent reads from its @c stdin stream.
+   *  end-of-file indicator on subsequent reads from its @c stdin stream.
    */
   template <typename C, typename T>
     inline void
@@ -1430,18 +1535,18 @@ namespace redi
     {
       sync();
       destroy_buffers(pstdin);
-      close_fd_array(&wpipe_, 1);
+      close_fd(wpipe_);
     }
 
   /**
+   * Unlike pstreambuf::exited(), this function will not call wait() and
+   * so will not change the object's state.  This means that once a child
+   * process is executed successfully this function will continue to
+   * return true even after the process exits (until wait() is called.)
+   *
    * @return  true if a previous call to open() succeeded and wait() has
    *          not been called and determined that the process has exited,
    *          false otherwise.
-   * @warning This function can not be used to determine whether the
-   *          command used to initialise the buffer was successfully
-   *          executed or not. If the shell command failed this function
-   *          will still return true.
-   *          You can use exited() to see if it's still open.
    */
   template <typename C, typename T>
     inline bool
@@ -1473,13 +1578,13 @@ namespace redi
 
   /**
    * Called when the internal character buffer is not present or is full,
-   * to transfer the buffer contents to the pipe. For unbuffered streams
-   * this is called for every insertion.
+   * to transfer the buffer contents to the pipe.
    *
    * @param   c  a character to be written to the pipe.
-   * @return  @c traits_type::not_eof(c) if @a c is equal to @c
-   *          traits_type::eof(). Otherwise returns @a c if @a c can be
-   *          written to the pipe, or @c traits_type::eof() if not.
+   * @return  @c traits_type::eof() if an error occurs, otherwise if @a c
+   *          is not equal to @c traits_type::eof() it will be buffered and
+   *          a value other than @c traits_type::eof() returned to indicate
+   *          success.
    */
   template <typename C, typename T>
     typename basic_pstreambuf<C,T>::int_type
@@ -1512,7 +1617,7 @@ namespace redi
     {
       if (n < this->epptr() - this->pptr())
       {
-        std::memcpy(this->pptr(), s, n * sizeof(char_type));
+        traits_type::copy(this->pptr(), s, n);
         this->pbump(n);
         return n;
       }
@@ -1535,11 +1640,16 @@ namespace redi
     basic_pstreambuf<C,T>::empty_buffer()
     {
       const std::streamsize count = this->pptr() - this->pbase();
-      const std::streamsize written = this->write(this->wbuffer_, count);
-      if (count > 0 && written == count)
+      if (count > 0)
       {
-        this->pbump(-written);
-        return true;
+        const std::streamsize written = this->write(this->wbuffer_, count);
+        if (written > 0)
+        {
+          if (const std::streamsize unwritten = count - written)
+            traits_type::move(this->pbase(), this->pbase()+written, unwritten);
+          this->pbump(-written);
+          return true;
+        }
       }
       return false;
     }
@@ -1550,7 +1660,6 @@ namespace redi
    *
    * @return The first available character in the buffer,
    * or @c traits_type::eof() in case of failure.
-   * @see uflow()
    */
   template <typename C, typename T>
     typename basic_pstreambuf<C,T>::int_type
@@ -1585,28 +1694,71 @@ namespace redi
          return traits_type::eof();
     }
 
+  template <typename C, typename T>
+    std::streamsize
+    basic_pstreambuf<C,T>::showmanyc()
+    {
+      int avail = 0;
+      if (sizeof(char_type) == 1)
+        avail = fill_buffer(true) ? this->egptr() - this->gptr() : -1;
+#ifdef FIONREAD
+      else
+      {
+        if (::ioctl(rpipe(), FIONREAD, &avail) == -1)
+          avail = -1;
+        else if (avail)
+          avail /= sizeof(char_type);
+      }
+#endif
+      return std::streamsize(avail);
+    }
+
   /**
    * @return  true if the buffer was filled, false otherwise.
    */
   template <typename C, typename T>
     bool
-    basic_pstreambuf<C,T>::fill_buffer()
+    basic_pstreambuf<C,T>::fill_buffer(bool non_blocking)
     {
       const std::streamsize pb1 = this->gptr() - this->eback();
       const std::streamsize pb2 = pbsz;
       const std::streamsize npb = std::min(pb1, pb2);
 
-      std::memmove( rbuffer() + pbsz - npb,
-                    this->gptr() - npb,
-                    npb * sizeof(char_type) );
+      char_type* const rbuf = rbuffer();
 
-      const std::streamsize rc = read(rbuffer() + pbsz, bufsz - pbsz);
+      traits_type::move(rbuf + pbsz - npb, this->gptr() - npb, npb);
 
-      if (rc > 0)
+      std::streamsize rc = -1;
+
+      if (non_blocking)
       {
-        this->setg( rbuffer() + pbsz - npb,
-                    rbuffer() + pbsz,
-                    rbuffer() + pbsz + rc );
+        const int flags = ::fcntl(rpipe(), F_GETFL);
+        if (flags != -1)
+        {
+          const bool blocking = !(flags & O_NONBLOCK);
+          if (blocking)
+            ::fcntl(rpipe(), F_SETFL, flags | O_NONBLOCK);  // set non-blocking
+
+          error_ = 0;
+          rc = read(rbuf + pbsz, bufsz - pbsz);
+
+          if (rc == -1 && error_ == EAGAIN)  // nothing available
+            rc = 0;
+          else if (rc == 0)  // EOF
+            rc = -1;
+
+          if (blocking)
+            ::fcntl(rpipe(), F_SETFL, flags); // restore
+        }
+      }
+      else
+        rc = read(rbuf + pbsz, bufsz - pbsz);
+
+      if (rc > 0 || (rc == 0 && non_blocking))
+      {
+        this->setg( rbuf + pbsz - npb,
+                    rbuf + pbsz,
+                    rbuf + pbsz + rc );
         return true;
       }
       else
@@ -1618,8 +1770,6 @@ namespace redi
 
   /**
    * Writes up to @a n characters to the pipe from the buffer @a s.
-   * This currently only works for fixed width character encodings where
-   * each character uses sizeof(char_type) bytes.
    *
    * @param   s  character buffer.
    * @param   n  buffer length.
@@ -1627,15 +1777,22 @@ namespace redi
    */
   template <typename C, typename T>
     inline std::streamsize
-    basic_pstreambuf<C,T>::write(char_type* s, std::streamsize n)
+    basic_pstreambuf<C,T>::write(const char_type* s, std::streamsize n)
     {
-      return wpipe() >= 0 ? ::write(wpipe(), s, n * sizeof(char_type)) : 0;
+      std::streamsize nwritten = 0;
+      if (wpipe() >= 0)
+      {
+        nwritten = ::write(wpipe(), s, n * sizeof(char_type));
+        if (nwritten == -1)
+          error_ = errno;
+        else
+          nwritten /= sizeof(char_type);
+      }
+      return nwritten;
     }
 
   /**
    * Reads up to @a n characters from the pipe to the buffer @a s.
-   * This currently only works for fixed width character encodings where
-   * each character uses sizeof(char_type) bytes.
    *
    * @param   s  character buffer.
    * @param   n  buffer length.
@@ -1645,7 +1802,16 @@ namespace redi
     inline std::streamsize
     basic_pstreambuf<C,T>::read(char_type* s, std::streamsize n)
     {
-      return rpipe() >= 0 ? ::read(rpipe(), s, n * sizeof(char_type)) : 0;
+      std::streamsize nread = 0;
+      if (rpipe() >= 0)
+      {
+        nread = ::read(rpipe(), s, n * sizeof(char_type));
+        if (nread == -1)
+          error_ = errno;
+        else
+          nread /= sizeof(char_type);
+      }
+      return nread;
     }
 
   /** @return a reference to the output file descriptor */
@@ -1690,7 +1856,7 @@ namespace redi
    * Abstract Base Class providing common functionality for basic_ipstream,
    * basic_opstream and basic_pstream.
    * pstream_common manages the basic_pstreambuf stream buffer that is used
-   * by the derived classes to initialise an IOStream class.
+   * by the derived classes to initialise an iostream class.
    */
 
   /** Creates an uninitialised stream. */
@@ -1848,9 +2014,10 @@ namespace redi
    * @warning  This function exposes the internals of the stream buffer and
    *           should be used with caution. It is the caller's responsibility
    *           to flush streams etc. in order to clear any buffered data.
-   *           The POSIX.1 function @c fdopen(3) is used to obtain the
+   *           The POSIX.1 function <b>fdopen</b>(3) is used to obtain the
    *           @c FILE pointers from the streambuf's private file descriptor
-   *           members so consult your system's documentation for @c fdopen().
+   *           members so consult your system's documentation for
+   *           <b>fdopen</b>(3).
    *
    * @param   in    A FILE* that will refer to the process' stdin.
    * @param   out   A FILE* that will refer to the process' stdout.
@@ -1862,10 +2029,12 @@ namespace redi
    * streams @c NULL is assigned to the parameter.
    * The return value can be tested to see which parameters should be
    * @c !NULL by masking with the corresponding @c pmode value.
+   *
+   * @see <b>fdopen</b>(3)
    */
   template <typename C, typename T>
     std::size_t
-    basic_pstreambuf<C,T>::fopen(std::FILE*& in, std::FILE*& out, std::FILE*& err)
+    basic_pstreambuf<C,T>::fopen(FILE*& in, FILE*& out, FILE*& err)
     {
       in = out = err = NULL;
       std::size_t open_files = 0;
@@ -1905,9 +2074,9 @@ namespace redi
    */
   template <typename C, typename T>
     inline std::size_t
-    pstream_common<C,T>::fopen(std::FILE*& in, std::FILE*& out, std::FILE*& err)
+    pstream_common<C,T>::fopen(FILE*& fin, FILE*& fout, FILE*& ferr)
     {
-      return buf_.fopen(in, out, err);
+      return buf_.fopen(fin, fout, ferr);
     }
 
 #endif // REDI_EVISCERATE_PSTREAMS
