@@ -1,13 +1,13 @@
 /*
 PStreams - POSIX Process I/O for C++
-Copyright (C) 2002 Jonathan Wakely
+Copyright (C) 2002-2008 Jonathan Wakely
 
 This file is part of PStreams.
 
 PStreams is free software; you can redistribute it and/or modify
-it under the terms of the GNU Lesser General Public License as
-published by the Free Software Foundation; either version 2.1 of
-the License, or (at your option) any later version.
+it under the terms of the GNU Lesser General Public License as published by
+the Free Software Foundation; either version 3 of the License, or
+(at your option) any later version.
 
 PStreams is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -15,10 +15,10 @@ MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.
 
 You should have received a copy of the GNU Lesser General Public License
-along with PStreams; if not, write to the Free Software Foundation, Inc.,
-59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
+#include <cassert>
 
 // TODO test rpstream more
 // TODO test whether error_ cleared after successful open().
@@ -40,16 +40,27 @@ along with PStreams; if not, write to the Free Software Foundation, Inc.,
 #include <string>
 #include <sstream>
 #include <fstream>
+#include <cstring>
 #include <unistd.h>
 #include <sys/types.h>
 #include <sys/stat.h>
 //#include <fcntl.h>
 #include <errno.h>
 
+
 #define PSTREAMS_VERSION_MAJOR PSTREAMS_VERSION & 0xff00
 #define PSTREAMS_VERSION_MINOR PSTREAMS_VERSION & 0x00f0
 #define PSTREAMS_VERSION_PATCHLEVEL PSTREAMS_VERSION & 0x000f
 
+#ifndef SLEEP_TIME
+// increase these if your OS takes a while for processes to exit
+# if defined (__sun) || defined(__APPLE__)
+#  define SLEEP_TIME 2
+# else
+#  define SLEEP_TIME 1
+# endif
+#endif
+
 using namespace std;
 using namespace redi;
 
@@ -62,7 +73,7 @@ template <>
     basic_pstreambuf<char>::sync()
     {
         std::cout.flush();  // makes terminated process clean up faster.
-        sleep(5);
+        sleep(SLEEP_TIME+3);
         std::cout.flush();  // makes terminated process clean up faster.
         return !exited() && empty_buffer() ? 0 : -1;
     }
@@ -79,31 +90,36 @@ template class redi::basic_rpstream<char>;
 
 namespace  // anon
 {
+    // process' exit status
+    int exit_status = 0;
+
     // helper functions for printing test results
 
-    char
-    test_type(istream const&)
-    { return 'r'; }
+    void
+    test_type(istream const&, char& c, int& i)
+    { static int count=0; c='r'; i=++count; }
 
-    char
-    test_type(ostream const&)
-    { return 'w'; }
+    void
+    test_type(ostream const&, char& c, int& i)
+    { static int count=0; c='w'; i=++count; }
 
-    char
-    test_type(iostream const&)
-    { return 'b'; }
+    void
+    test_type(iostream const&, char& c, int& i)
+    { static int count=0; c='b'; i=++count; }
 
-    char
-    test_type(rpstream const&)
-    { return 'x'; }
+    void
+    test_type(rpstream const&, char& c, int& i)
+    { static int count=0; c='x'; i=++count; }
 
     template <typename T>
     string
     test_id(T const& s)
     {
-        static int count = 0;
+        char label = '?';
+        int count = 0;
+        test_type(s, label, count);
         ostringstream buf;
-        buf << test_type(s) << ++count;
+        buf << label << count;
         return buf.str();
     }
 
@@ -114,6 +130,7 @@ namespace  // anon
         clog << "Test " << setw(4) << test_id(s) << ": "
             << (result ? "Pass" : "Fail!")
             << endl;
+        exit_status += !result;
     }
 
     template <typename T>
@@ -137,11 +154,16 @@ namespace  // anon
 #endif
 }
 
+sig_atomic_t sig_counter = 0;
+extern "C" void my_sig_handler(int) { ++sig_counter; }
 
 int main()
 {
     ios_base::sync_with_stdio();
 
+    const pstreams::pmode all3streams =
+        pstreams::pstdin|pstreams::pstdout|pstreams::pstderr;
+
     string str;
 
     clog << "# Testing basic I/O\n";
@@ -248,18 +270,11 @@ int main()
 
     clog << "# Testing bidirectional PStreams\n";
 
-    const pstreams::pmode all3streams =
-        pstreams::pstdin|pstreams::pstdout|pstreams::pstderr;
-
     {
         // test reading from bidirectional pstream
 
-#if defined(__sun)
-        // Solaris' grep doesn't like "--" and "-"
         const string cmd = "grep '^127' /etc/hosts /no/such/file /dev/stdin";
-#else
-        const string cmd = "grep '^127' -- /etc/hosts /no/such/file -";
-#endif
+
         pstream ps(cmd, all3streams);
 
         print_result(ps, ps.is_open());
@@ -283,12 +298,8 @@ int main()
         // test input on bidirectional pstream
         // and test child moves onto next file after peof on stdin
 
-#if defined (__sun)
-        // Solaris' grep doesn't like "--" and "-"
         const string cmd = "grep fnord /etc/hosts /dev/stdin";
-#else
-        const string cmd = "grep fnord -- /etc/hosts -";
-#endif
+
         pstream ps(cmd, all3streams);
 
         print_result(ps, ps.is_open());
@@ -324,7 +335,7 @@ int main()
         const int e2 = pbuf->error();
         print_result(ps, e1 == e2);
 
-        sleep(3);  // allow time for child process to exit completely
+        sleep(SLEEP_TIME);  // allow time for child process to exit completely
 
         // close() will call sync(), which shouldn't flush buffer after kill()
         pbuf->close();
@@ -344,7 +355,7 @@ int main()
         check_pass(ps.err());
 
         ps.rdbuf()->kill();
-        ::sleep(3);
+        ::sleep(SLEEP_TIME);
         print_result(ps, ps.is_open());
         print_result(ps, ps.rdbuf()->exited());
         print_result(ps, !ps.is_open());
@@ -374,7 +385,11 @@ int main()
         print_result(ps, !pbuf->exited());
         ps.close();
         print_result(ps, pbuf->exited());
+        check_pass(ps);
+        ps.close();  // should set failbit on second call:
+        check_fail(ps);
 
+        ps.clear();
         ps.open(cmd);
         print_result(ps, ps.is_open());
         print_result(ps, !pbuf->exited());
@@ -382,7 +397,6 @@ int main()
         print_result(ps, pbuf->exited());
     }
 
-
     clog << "# Testing behaviour with bad commands" << endl;
 
     //string badcmd = "hgfhdgf";
@@ -392,10 +406,11 @@ int main()
         // check is_open() works 
         ipstream is(badcmd);
         // print_result(is, !is.is_open());  // XXX cannot pass this test!
-#if defined (__sun) || defined(__APPLE__)
-        // fail next test if OS slow to terminate child process, need sleep(1)
-        sleep(1);
-#endif
+
+        (void) is.err();
+        while (!is.rdbuf()->in_avail())  // need to wait for exit
+            ::sleep(1);
+
         print_result(is, is.rdbuf()->exited() && !is.is_open());
     }
 
@@ -419,9 +434,7 @@ int main()
     {
         // test writing to bad command
         opstream ofail(badcmd);
-#if defined (__sun) || defined(__APPLE__)
-        sleep(1);  // give shell time to try command and exit
-#endif
+        ::sleep(SLEEP_TIME);
         // this would cause SIGPIPE: ofail<<"blahblah";
         // does not show failure: print_result(ofail, !ofail.is_open());
         pstreambuf* buf = ofail.rdbuf();
@@ -488,10 +501,39 @@ int main()
 
     {
         // testing streambuf::in_avail()
-        ipstream in("hostname");
+        ipstream in("{ printf 'this is ' ; sleep 2 ; printf 'hardcore' ; }");
+        sleep(1);
         streamsize avail = in.rdbuf()->in_avail();
-        cout << "STDOUT: " << avail << " characters: " << in.rdbuf();
         print_result(in, avail > 0);
+        print_result(in, size_t(avail) == strlen("this is "));
+
+        std::vector<char> buf;
+        int tries = 0;
+        while (in && avail >= 0)
+        {
+            if (avail > 0)
+            {
+                if (tries)
+                {
+                    cout << "Nothing to read " << tries << " times." << endl;
+                    tries = 0;
+                }
+                buf.resize(avail);
+                in.readsome(&buf.front(), avail);
+                cout << "STDOUT: " << avail << " characters: "
+                    << string(buf.begin(), buf.end()) << endl;
+            }
+            else
+                ++tries;
+            avail = in.rdbuf()->in_avail();
+        }
+
+        if (tries)
+            cout << "Nothing to read " << tries << " times." << endl;
+        char c;
+        check_pass(in);
+        check_fail(in >> c);
+        print_result(in, in.eof());
     }
 
     // TODO more testing of other members
@@ -569,6 +611,35 @@ int main()
         check_fail(in.err().get(c));
     }
 
+    clog << "# Testing initial pmode set correctly\n";
+    {
+        char c;
+        string s;
+        ipstream in("echo 'abc' >&2", pstreambuf::pstderr);
+        print_result(in, getline(in, s));
+        print_result(in, s == "abc");
+        check_fail(in.get(c));
+        in.close();
+        in.clear(); // clear EOF
+        s.erase();
+
+        in.open("echo 'abc'", pstreambuf::pstdout);
+        print_result(in, getline(in, s));
+        print_result(in, s == "abc");
+        check_fail(in.get(c));
+        in.close();
+        in.clear(); // clear EOF
+        s.erase();
+
+        in.open("echo 'abc' >&2", pstreambuf::pstderr);
+        print_result(in, getline(in, s));
+        print_result(in, s == "abc");
+        check_fail(in.get(c));
+        in.close();
+        in.clear(); // clear EOF
+        s.erase();
+    }
+
 #if REDI_EVISCERATE_PSTREAMS
     clog << "# Testing eviscerated pstream\n";
 
@@ -633,51 +704,148 @@ int main()
 #endif
 
     clog << "# Testing resources freed correctly\n";
-    // TODO repeat tests for vector open()
-
-    {
-        const int next_fd = dup(0);
-        ::close(next_fd);
-
-        ipstream in("hostname");
-
-        ::sleep(3);  // wait for process to exit
-
-        in.rdbuf()->exited();  // test for exit, destroy buffers
-
-        // check no open files except for stdin, stdout, stderr
-        int fd = dup(0);
-        print_result(in, next_fd == fd);
-        ::close(fd);
-    }
 
     {
         const int next_fd = dup(0);
         ::close(next_fd);
 
         pstream p("cat", all3streams);
+        p << "fnord" << peof;
 
-        ::sleep(3);  // wait for process to exit
-        p.rdbuf()->exited();  // test for exit, destroy buffers
+        ::sleep(SLEEP_TIME);  // wait for process to exit
+
+        p.rdbuf()->exited();  // test for exit, destroys write buffer
+
+        // check we can still read the output
+        std::string s;
+        print_result(p,  p >> s || p.err() >> s);
+
+        p.close(); // destroys read buffers, closes all pipes
+
+        check_fail(p >> s);
 
         // check no open files except for stdin, stdout, stderr
         int fd = dup(0);
         print_result(p, next_fd == fd);
         ::close(fd);
 
-        // now close and reopen and check again
+        // reopen and check again with vector-open
+
+        std::vector<std::string> argv;
+        argv.push_back("cat");
+        p.open(argv[0], argv, all3streams);
 
         p.close();
 
-        p.open("cat", all3streams);
-        ::sleep(3);  // wait for process to exit
-        p.rdbuf()->exited();  // test for exit, destroy buffers
+        // reopen and check again
+
+        p.open(argv[0], all3streams);
+        p.close();
 
         fd = dup(0);
         print_result(p, next_fd == fd);
         ::close(fd);
     }
 
-    return 0;
+    {
+        // let's see if we can leak in the face of signals...
+
+        const int next_fd = dup(0);
+        ::close(next_fd);
+
+        typedef struct sigaction sa;
+        sa act = sa(), oldact;
+        act.sa_handler = my_sig_handler;
+        sigaction(SIGALRM, &act, &oldact);
+
+        std::vector<std::string> argv;
+        argv.push_back("sleep 5");
+
+        ipstream in(argv[0]);
+        alarm(3);
+        // close() should hang until interrupted then finish closing
+        in.close();
+
+        // check no open files except for stdin, stdout, stderr
+        int fd = dup(0);
+        print_result(in, fd == next_fd);
+        ::close(fd);
+
+        in.open(argv[0], argv);
+        alarm(3);
+        in.close();
+
+        sigaction(SIGALRM, &oldact, 0);
+
+        // check no open files except for stdin, stdout, stderr
+        fd = dup(0);
+        print_result(in, fd == next_fd);
+        ::close(fd);
+
+        class pguard
+        {
+        public:
+            explicit pguard(ipstream& in, int signal=SIGKILL)
+            : buf_(*in.rdbuf()), signal_(signal) { }
+
+          ~pguard() { if (signal_) buf_.kill(signal_); }
+
+          void release() { signal_ = 0; }
+
+        private:
+          pstreambuf& buf_;
+          int signal_;
+        };
+
+        in.clear();
+        in.open("echo sleeping && sleep 5");
+        std::string s;
+        print_result(in, in.is_open() && in >> s);
+
+        sigaction(SIGALRM, &act, &oldact);
+        sig_counter = 0;
+        alarm(3);
+
+        try
+        {
+            pguard pg(in, SIGTERM);  // should kill child
+            throw 0;
+        } catch(...) {
+            in.close();  // should return immediately if child killed
+            print_result(in, alarm(0) > 0 && sig_counter == 0);
+            sigaction(SIGALRM, &oldact, 0);
+
+            int status = in.rdbuf()->status();
+            print_result(in, WIFSIGNALED(status) && WTERMSIG(status)==SIGTERM);
+        }
+    }
+
+    clog << "# Testing wide chars\n";
+    {
+        ipstream dummy("true");
+        basic_ipstream<wchar_t> in("cat ./pstreams.wout");
+        wstring s;
+        in >> s;
+        wcout << s;
+        print_result(dummy, in);
+
+        wchar_t wc;
+        int count=0, gcount=0;
+        while (in.get(wc))
+        {
+            wcout << wc;
+            ++count;
+            gcount += in.gcount();
+        }
+        wcout << L'\n';
+        print_result(dummy, in.eof() && in.fail());
+        print_result(dummy, gcount == count);
+        wcout << L"Read: " << gcount << L" chars." << endl;
+    }
+
+
+
+
+    return exit_status;
 }
 
